/*	$NetBSD: quota.c,v 1.23 2001/03/29 21:54:01 cgd Exp $	*/

/*
 * Copyright (c) 1980, 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Robert Elz at The University of Melbourne.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * NFS Disk quota reporting program.
 */
#include <sys/param.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/socket.h>

#include <ctype.h>
#include <errno.h>
#include <grp.h>
#include <limits.h>
#include <netdb.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <sys/mnttab.h>

#include <rpc/rpc.h>
#include <rpc/pmap_prot.h>
#include "rquota.h"

/* From NetBSD <ufs/ufs/quota.h> */
struct dqblk {
	uint32_t dqb_bhardlimit;       /* absolute limit on disk blks alloc */
	uint32_t dqb_bsoftlimit;       /* preferred limit on disk blks */
	uint32_t dqb_curblocks;	/* current block count */
	uint32_t dqb_ihardlimit;       /* maximum # allocated inodes + 1 */
	uint32_t dqb_isoftlimit;       /* preferred inode limit */
	uint32_t dqb_curinodes;	/* current # allocated inodes */
	int32_t   dqb_btime;		/* time limit for excessive disk use */
	int32_t   dqb_itime;		/* time limit for excessive files */
};


struct quotause {
	struct	quotause *next;
	long	flags;
	struct	dqblk dqblk;
	char	fsname[MAXPATHLEN + 1];
};
#define	FOUND	0x01

char *qfextension[] = {"user", "group"};

int	alldigits (char *);
int	callaurpc (char *, int, int, int, xdrproc_t, void *,
	    xdrproc_t, void *);
int	main (int, char **);
int	getnfsquota (struct mnttab *, struct quotause *, long, int);
struct quotause	*getprivs (long id, int quotatype);
void	heading (int, u_long, const char *, const char *);
void	showgid (gid_t);
void	showgrpname (const char *);
void	showquotas (int, u_long, const char *);
void	showuid (uid_t);
void	showusrname (const char *);
char   *timeprt (time_t seconds);
void	usage (void);

int	qflag;
int	vflag;
uid_t	myuid;

char *progname;

int
main(argc, argv)
	int argc;
	char *argv[];
{
	int ngroups; 
	gid_t mygid, gidset[NGROUPS_MAX];
	int i, gflag = 0, uflag = 0;
	int ch;

	progname = argv[0];

	myuid = getuid();
	while ((ch = getopt(argc, argv, "ugvq")) != -1) {
		switch(ch) {
		case 'g':
			gflag++;
			break;
		case 'u':
			uflag++;
			break;
		case 'v':
			vflag++;
			break;
		case 'q':
			qflag++;
			break;
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;
	if (!uflag && !gflag)
		uflag++;
	if (argc == 0) {
		if (uflag)
			showuid(myuid);
		if (gflag) {
			mygid = getgid();
			ngroups = getgroups(NGROUPS_MAX, gidset);
			if (ngroups < 0) {
				fprintf(stderr, "%s: getgroups\n", progname);
				exit(1);
			}
			showgid(mygid);
			for (i = 0; i < ngroups; i++)
				if (gidset[i] != mygid)
					showgid(gidset[i]);
		}
		exit(0);
	}
	if (uflag && gflag)
		usage();
	if (uflag) {
		for (; argc > 0; argc--, argv++) {
			if (alldigits(*argv))
				showuid(atoi(*argv));
			else
				showusrname(*argv);
		}
		exit(0);
	}
	if (gflag) {
		for (; argc > 0; argc--, argv++) {
			if (alldigits(*argv))
				showgid(atoi(*argv));
			else
				showgrpname(*argv);
		}
		exit(0);
	}
	/* NOTREACHED */
	return (0);
}

void
usage()
{

	fprintf(stderr, "%s\n%s\n%s\n",
	    "Usage: quota [-guqv]",
	    "\tquota [-qv] -u username ...",
	    "\tquota [-qv] -g groupname ...");
	exit(1);
}

/*
 * Print out quotas for a specified user identifier.
 */
void
showuid(uid)
	uid_t uid;
{
	struct passwd *pwd = getpwuid(uid);
	const char *name;

	if (pwd == NULL)
		name = "(no account)";
	else
		name = pwd->pw_name;
	if (uid != myuid && myuid != 0) {
		printf("quota: %s (uid %d): permission denied\n", name, uid);
		return;
	}
	showquotas(RQUOTA_USRQUOTA, uid, name);
}

/*
 * Print out quotas for a specifed user name.
 */
void
showusrname(name)
	const char *name;
{
	struct passwd *pwd = getpwnam(name);

	if (pwd == NULL) {
		fprintf(stderr, "%s: %s: unknown user\n", progname, name);
		return;
	}
	if (pwd->pw_uid != myuid && myuid != 0) {
		fprintf(stderr, "%s: %s (uid %d): permission denied\n",
		    progname, name, pwd->pw_uid);
		return;
	}
	showquotas(RQUOTA_USRQUOTA, pwd->pw_uid, name);
}

/*
 * Print out quotas for a specified group identifier.
 */
void
showgid(gid)
	gid_t gid;
{
	struct group *grp = getgrgid(gid);
	int ngroups;
	gid_t mygid, gidset[NGROUPS_MAX];
	int i;
	const char *name;

	if (grp == NULL)
		name = "(no entry)";
	else
		name = grp->gr_name;
	mygid = getgid();
	ngroups = getgroups(NGROUPS_MAX, gidset);
	if (ngroups < 0) {
		fprintf(stderr, "%s: getgroups\n", progname);
		return;
	}
	if (gid != mygid) {
		for (i = 0; i < ngroups; i++)
			if (gid == gidset[i])
				break;
		if (i >= ngroups && myuid != 0) {
			fprintf(stderr, "%s: %s (gid %d): permission denied\n",
			    progname, name, gid);
			return;
		}
	}
	showquotas(RQUOTA_GRPQUOTA, gid, name);
}

/*
 * Print out quotas for a specifed group name.
 */
void
showgrpname(name)
	const char *name;
{
	struct group *grp = getgrnam(name);
	int ngroups;
	gid_t mygid, gidset[NGROUPS_MAX];
	int i;

	if (grp == NULL) {
		fprintf(stderr, "%s: %s: unknown group\n", progname, name);
		return;
	}
	mygid = getgid();
	ngroups = getgroups(NGROUPS_MAX, gidset);
	if (ngroups < 0) {
		fprintf(stderr, "%s: getgroups\n", progname);
		return;
	}
	if (grp->gr_gid != mygid) {
		for (i = 0; i < ngroups; i++)
			if (grp->gr_gid == gidset[i])
				break;
		if (i >= ngroups && myuid != 0) {
			fprintf(stderr, "%s: %s (gid %d): permission denied\n",
			    progname, name, grp->gr_gid);
			return;
		}
	}
	showquotas(RQUOTA_GRPQUOTA, grp->gr_gid, name);
}

void
showquotas(type, id, name)
	int type;
	u_long id;
	const char *name;
{
	struct quotause *qup;
	struct quotause *quplist;
	char *msgi, *msgb, *nam;
	int lines = 0;
	static time_t now;

	if (now == 0)
		time(&now);
	quplist = getprivs(id, type);
	for (qup = quplist; qup; qup = qup->next) {
		if (!vflag &&
		    qup->dqblk.dqb_isoftlimit == 0 &&
		    qup->dqblk.dqb_ihardlimit == 0 &&
		    qup->dqblk.dqb_bsoftlimit == 0 &&
		    qup->dqblk.dqb_bhardlimit == 0)
			continue;
		msgi = (char *)0;
		if (qup->dqblk.dqb_ihardlimit &&
		    qup->dqblk.dqb_curinodes >= qup->dqblk.dqb_ihardlimit)
			msgi = "File limit reached on";
		else if (qup->dqblk.dqb_isoftlimit &&
		    qup->dqblk.dqb_curinodes >= qup->dqblk.dqb_isoftlimit) {
			if (qup->dqblk.dqb_itime > now)
				msgi = "In file grace period on";
			else
				msgi = "Over file quota on";
		}
		msgb = (char *)0;
		if (qup->dqblk.dqb_bhardlimit &&
		    qup->dqblk.dqb_curblocks >= qup->dqblk.dqb_bhardlimit)
			msgb = "Block limit reached on";
		else {
			if (qup->dqblk.dqb_bsoftlimit
			&&  qup->dqblk.dqb_curblocks
			    >= qup->dqblk.dqb_bsoftlimit) {
				if (qup->dqblk.dqb_btime > now)
					msgb = "In block grace period on";
				else
					msgb = "Over block quota on";
			}
		}
		if (qflag) {
			if ((msgi != (char *)0 || msgb != (char *)0) &&
			    lines++ == 0)
				heading(type, id, name, "");
			if (msgi != (char *)0)
				printf("\t%s %s\n", msgi, qup->fsname);
			if (msgb != (char *)0)
				printf("\t%s %s\n", msgb, qup->fsname);
			continue;
		}
		if (vflag ||
		    qup->dqblk.dqb_curblocks ||
		    qup->dqblk.dqb_curinodes) {
			if (lines++ == 0)
				heading(type, id, name, "");
			nam = qup->fsname;
			if (strlen(qup->fsname) > 15) {
				printf("%s\n", qup->fsname);
				nam = "";
			} 
#ifndef __svr4__
			printf("%12s%9d%c%8d%9d%8s"
			    , nam
			    , (int)(dbtob((uint64_t)qup->dqblk.dqb_curblocks)
				/ 1024)
			    , (msgb == (char *)0) ? ' ' : '*'
			    , (int)(dbtob((uint64_t)qup->dqblk.dqb_bsoftlimit)
				/ 1024)
			    , (int)(dbtob((uint64_t)qup->dqblk.dqb_bhardlimit)
				/ 1024)
			    , (msgb == (char *)0) ? ""
			        : timeprt(qup->dqblk.dqb_btime));
#else
			printf("%12s%9d%c%8d%9d%8s"
			    , nam
			    , (int)(ldbtob((uint64_t)qup->dqblk.dqb_curblocks)
				/ 1024)
			    , (msgb == (char *)0) ? ' ' : '*'
			    , (int)(ldbtob((uint64_t)qup->dqblk.dqb_bsoftlimit)
				/ 1024)
			    , (int)(ldbtob((uint64_t)qup->dqblk.dqb_bhardlimit)
				/ 1024)
			    , (msgb == (char *)0) ? ""
			        : timeprt(qup->dqblk.dqb_btime));
#endif
			printf("%8d%c%7d%8d%8s\n"
			    , qup->dqblk.dqb_curinodes
			    , (msgi == (char *)0) ? ' ' : '*'
			    , qup->dqblk.dqb_isoftlimit
			    , qup->dqblk.dqb_ihardlimit
			    , (msgi == (char *)0) ? ""
			        : timeprt(qup->dqblk.dqb_itime)
			);
			continue;
		}
	}
	if (!qflag && lines == 0)
		heading(type, id, name, "none");
}

void
heading(type, id, name, tag)
	int type;
	u_long id;
	const char *name, *tag;
{

	printf("Disk quotas for %s %s (%cid %ld): %s\n", qfextension[type],
	    name, *qfextension[type], (u_long)id, tag);
	if (!qflag && tag[0] == '\0') {
		printf("%12s%9s %8s%9s%8s%8s %7s%8s%8s\n"
		    , "Filesystem"
		    , "blocks"
		    , "quota"
		    , "limit"
		    , "grace"
		    , "files"
		    , "quota"
		    , "limit"
		    , "grace"
		);
	}
}

/*
 * Calculate the grace period and return a printable string for it.
 */
char *
timeprt(seconds)
	time_t seconds;
{
	time_t hours, minutes;
	static char buf[20];
	static time_t now;

	if (now == 0)
		time(&now);
	if (now > seconds)
		return ("none");
	seconds -= now;
	minutes = (seconds + 30) / 60;
	hours = (minutes + 30) / 60;
	if (hours >= 36) {
		(void)snprintf(buf, sizeof buf, "%ddays",
		    (int)((hours + 12) / 24));
		return (buf);
	}
	if (minutes >= 60) {
		(void)snprintf(buf, sizeof buf, "%2d:%d",
		    (int)(minutes / 60), (int)(minutes % 60));
		return (buf);
	}
	(void)snprintf(buf, sizeof buf, "%2d", (int)minutes);
	return (buf);
}

/*
 * Collect the requested quota information.
 */
struct quotause *
getprivs(id, quotatype)
	long id;
	int quotatype;
{
	struct quotause *qup, *quptail;
	struct mnttab fs;
	struct quotause *quphead;
	int i;
	FILE *mt;

	qup = quphead = quptail = NULL;

	mt = fopen("/etc/mnttab", "r");
	if (mt == NULL) {
		fprintf(stderr, "%s: open /etc/mnttab: %s\n", progname,
		    strerror(errno));
		exit(1);
	}
	while (getmntent(mt, &fs) == 0) {
		if (qup == NULL) {
			if ((qup =
			    (struct quotause *)malloc(sizeof *qup)) == NULL) {
				fprintf(stderr, "%s: out of memory\n",
				    progname);
				exit(2);
			}
		}
		if (strcmp(fs.mnt_fstype, "nfs") == 0) {
			if (getnfsquota(&fs, qup, id, quotatype) == 0)
				continue;
		} else
			continue;
		(void)strncpy(qup->fsname, fs.mnt_mountp,
		    sizeof(qup->fsname) - 1);
		if (quphead == NULL)
			quphead = qup;
		else
			quptail->next = qup;
		quptail = qup;
		quptail->next = 0;
		qup = NULL;
	}
	if (qup)
		free(qup);
	return (quphead);
}

int
getnfsquota(fs, qup, id, quotatype)
	struct mnttab *fs;
	struct quotause *qup;
	long id;
	int quotatype;
{
	struct getquota_args gq_args;
	struct ext_getquota_args ext_gq_args;
	struct getquota_rslt gq_rslt;
	struct dqblk *dqp = &qup->dqblk;
	struct timeval tv;
	char *cp;
	int ret;

	/*
	 * must be some form of "hostname:/path"
	 */
	cp = strchr(fs->mnt_special, ':');
	if (cp == NULL) {
		fprintf(stderr, "%s: cannot find hostname for %s\n",
		    progname, fs->mnt_special);
		return (0);
	}
 
	*cp = '\0';
	if (*(cp+1) != '/') {
		*cp = ':';
		return (0);
	}

	ext_gq_args.gqa_pathp = cp + 1;
	ext_gq_args.gqa_id = id;
	ext_gq_args.gqa_type = quotatype;
	ret = callaurpc(fs->mnt_special, RQUOTAPROG, EXT_RQUOTAVERS,
	    RQUOTAPROC_GETQUOTA, xdr_ext_getquota_args, &ext_gq_args,
	    xdr_getquota_rslt, &gq_rslt);
	if (ret == RPC_VERSMISMATCH) {
		if (quotatype != RQUOTA_USRQUOTA) {
			*cp = ':';
			return (0);
		}
		/* try RQUOTAVERS */
		gq_args.gqa_pathp = cp + 1;
		gq_args.gqa_uid = id;
		ret = callaurpc(fs->mnt_special, RQUOTAPROG, RQUOTAVERS,
		    RQUOTAPROC_GETQUOTA, xdr_getquota_args, &gq_args,
			    xdr_getquota_rslt, &gq_rslt);
	}
	if (ret != RPC_SUCCESS) {
		*cp = ':';
		return (0);
	}

	switch (gq_rslt.status) {
	case Q_NOQUOTA:
		break;
	case Q_EPERM:
		fprintf(stderr, "%s: quota permission error, host: %s\n",
		    progname, fs->mnt_special);
		break;
	case Q_OK:
		gettimeofday(&tv, NULL);
			/* blocks*/
		dqp->dqb_bhardlimit =
		    gq_rslt.getquota_rslt_u.gqr_rquota.rq_bhardlimit *
		    (gq_rslt.getquota_rslt_u.gqr_rquota.rq_bsize / DEV_BSIZE);
		dqp->dqb_bsoftlimit =
		    gq_rslt.getquota_rslt_u.gqr_rquota.rq_bsoftlimit *
		    (gq_rslt.getquota_rslt_u.gqr_rquota.rq_bsize / DEV_BSIZE);
		dqp->dqb_curblocks =
		    gq_rslt.getquota_rslt_u.gqr_rquota.rq_curblocks *
		    (gq_rslt.getquota_rslt_u.gqr_rquota.rq_bsize / DEV_BSIZE);
			/* inodes */
		dqp->dqb_ihardlimit =
			gq_rslt.getquota_rslt_u.gqr_rquota.rq_fhardlimit;
		dqp->dqb_isoftlimit =
			gq_rslt.getquota_rslt_u.gqr_rquota.rq_fsoftlimit;
		dqp->dqb_curinodes =
			gq_rslt.getquota_rslt_u.gqr_rquota.rq_curfiles;
			/* grace times */
		dqp->dqb_btime =
		    tv.tv_sec + gq_rslt.getquota_rslt_u.gqr_rquota.rq_btimeleft;
		dqp->dqb_itime =
		    tv.tv_sec + gq_rslt.getquota_rslt_u.gqr_rquota.rq_ftimeleft;
		*cp = ':';
		return (1);
	default:
		fprintf(stderr, "%s: bad rpc result, host: %s\n",
		    progname, fs->mnt_special);
		break;
	}
	*cp = ':';
	return (0);
}
 
int
callaurpc(host, prognum, versnum, procnum, inproc, in, outproc, out)
	char *host;
	int prognum, versnum, procnum;
	xdrproc_t inproc;
	void *in;
	xdrproc_t outproc;
	void *out;
{
	struct sockaddr_in server_addr;
	enum clnt_stat clnt_stat;
	struct hostent *hp;
	struct timeval timeout, tottimeout;
 
	CLIENT *client = NULL;
	int socket = RPC_ANYSOCK;
 
	if ((hp = gethostbyname(host)) == NULL)
		return ((int) RPC_UNKNOWNHOST);
	timeout.tv_usec = 0;
	timeout.tv_sec = 6;
	memmove(&server_addr.sin_addr, hp->h_addr, hp->h_length);
	server_addr.sin_family = AF_INET;
	server_addr.sin_port =  0;

	if ((client = clntudp_create(&server_addr, prognum,
	    versnum, timeout, &socket)) == NULL)
		return ((int) rpc_createerr.cf_stat);

	client->cl_auth = authunix_create_default();
	tottimeout.tv_sec = 25;
	tottimeout.tv_usec = 0;
	clnt_stat = clnt_call(client, procnum, inproc, in,
	    outproc, out, tottimeout);
 
	return ((int) clnt_stat);
}

int
alldigits(s)
	char *s;
{
	int c;

	c = *s++;
	do {
		if (!isdigit(c))
			return (0);
	} while ((c = *s++) != 0);
	return (1);
}
